@@ -23,7 +23,7 @@ import logging
 from collections.abc import Iterable
 from datetime import date, datetime, time, timedelta, timezone
 from collections.abc import Callable
-from typing import Protocol
+from typing import Protocol, overload
 from zoneinfo import ZoneInfo
 
 import dateutil.rrule
@@ -36,6 +36,7 @@ from icalendar.prop import (
     vDatetime,
     vDDDTypes,
     vDuration,
+    vRecur,
     vText,
 )
 
@@ -62,22 +63,25 @@ MAX_EXPANSION_TIME = datetime(
     9999, 12, 31, 23, 59, 59, tzinfo=timezone.utc
 )  # 99991231T235959Z
 
+# Maximum number of recurrence instances to expand for infinite recurrences
+# This prevents resource exhaustion when expanding recurring events with no end date
+# Following sabre/dav and Stalwart's approach of limiting to ~3000 instances
+MAX_RECURRENCE_INSTANCES = 3000
+
 
 # Based on RFC5545 section 3.3.11, CONTROL = %x00-08 / %x0A-1F / %x7F
-# All control characters except HTAB (\x09) are forbidden
-_INVALID_CONTROL_CHARACTERS = (
-    [
-        chr(i)
-        for i in range(0x00, 0x09)  # \x00-\x08
-    ]
-    + [
-        chr(i)
-        for i in range(0x0A, 0x20)  # \x0A-\x1F
-    ]
-    + [
-        chr(0x7F)  # DEL character
-    ]
-)
+# Control characters are forbidden in TEXT values, EXCEPT:
+# - HTAB (\x09) is explicitly allowed
+# - LF (\x0A) and CR (\x0D) are allowed because they appear in the parsed
+#   representation when the icalendar library unescapes valid \n and \r
+#   escape sequences from the iCalendar file (RFC 5545 allows these escapes)
+_INVALID_CONTROL_CHARACTERS = [
+    chr(i)
+    for i in range(0x00, 0x20)  # \x00-\x1F
+    if i not in (0x09, 0x0A, 0x0D)  # Allow HTAB, LF, CR
+] + [
+    chr(0x7F)  # DEL character
+]
 
 
 class MissingProperty(Exception):
@@ -434,8 +438,17 @@ def apply_time_range_vfreebusy(start, end, comp, tzify):
 def _create_enriched_valarm(alarm: Component, parent: Component) -> Component:
     """Create a modified VALARM component with calculated absolute trigger time.
 
-    This creates a new VALARM that converts relative triggers to absolute
-    triggers based on the parent component's timing properties.
+    This creates a new VALARM that converts relative triggers (e.g., "-PT15M")
+    to absolute trigger times based on the parent component's DTSTART/DUE properties.
+    This enrichment is necessary for time-range filtering to work correctly with
+    relative triggers, and must be applied consistently in both filtering and indexing.
+
+    Args:
+        alarm: The VALARM component to enrich
+        parent: The parent component (VEVENT/VTODO) containing timing information
+
+    Returns:
+        A new VALARM component with absolute TRIGGER if applicable, otherwise unchanged
     """
     from icalendar.cal import Alarm
 
@@ -610,7 +623,16 @@ class PropertyTimeRangeMatcher:
         dt = tzify(prop.dt)
         return dt >= self.start and dt <= self.end
 
-    def match_indexes(self, prop: SubIndexDict, tzify: TzifyFunction):
+    def match_indexes(
+        self, prop: SubIndexDict, tzify: TzifyFunction, context: str | None = None
+    ):
+        """Match indexes against this property time range matcher.
+
+        Args:
+            prop: Property index dictionary to match against
+            tzify: Timezone conversion function
+            context: Optional context string (e.g. filename) for error/warning reporting (unused here)
+        """
         return any(
             self.match(vDDDTypes(vDDDTypes.from_ical(p.decode("utf-8"))), tzify)
             for p in prop[None]
@@ -665,11 +687,20 @@ class ComponentTimeRangeMatcher:
             return False
         return component_handler(self.start, self.end, comp, tzify)
 
-    def match_indexes(self, indexes: SubIndexDict, tzify: TzifyFunction):
+    def match_indexes(
+        self, indexes: SubIndexDict, tzify: TzifyFunction, context: str | None = None
+    ):
+        """Match indexes against this time range matcher.
+
+        Args:
+            indexes: Index dictionary to match against
+            tzify: Timezone conversion function
+            context: Optional context string (e.g. filename) for error/warning reporting
+        """
         # Check if we have RRULE - if so, expand and test occurrences
         rrule_values = indexes.get("P=RRULE")
         if rrule_values and rrule_values[0]:
-            return self._match_indexes_with_rrule(indexes, tzify)
+            return self._match_indexes_with_rrule(indexes, tzify, context)
 
         # Original logic for non-recurring events
         vs: list[dict[str, vDDDTypes | None]] = []
@@ -726,8 +757,16 @@ class ComponentTimeRangeMatcher:
                 return True
         return False
 
-    def _match_indexes_with_rrule(self, indexes: SubIndexDict, tzify: TzifyFunction):
-        """Handle time-range matching for recurring events using RRULE expansion."""
+    def _match_indexes_with_rrule(
+        self, indexes: SubIndexDict, tzify: TzifyFunction, context: str | None = None
+    ):
+        """Handle time-range matching for recurring events using RRULE expansion.
+
+        Args:
+            indexes: Index dictionary to match against
+            tzify: Timezone conversion function
+            context: Optional context string (e.g. filename) for error/warning reporting
+        """
         # Extract and validate RRULE and DTSTART from indexes
         rrule_values = indexes.get("P=RRULE", [])
         dtstart_values = indexes.get("P=DTSTART", [])
@@ -757,10 +796,29 @@ class ComponentTimeRangeMatcher:
 
             # Parse DTSTART and create rrule
             dtstart_parsed = vDDDTypes.from_ical(dtstart_str)
+            # Normalize UNTIL to be UTC if dtstart is timezone-aware
+            rrule_str = _normalize_rrule_until(rrule_str, dtstart_parsed)
             rrule = dateutil.rrule.rrulestr(rrule_str, dtstart=dtstart_parsed)
         except (TypeError, ValueError) as e:
-            # If RRULE parsing fails, log and return False
-            logging.warning("Failed to parse RRULE in time-range filter: %s", e)
+            # If RRULE parsing fails, log with context and return False
+            uid_values = indexes.get("P=UID", [])
+            uid = (
+                decode_bytes(uid_values[0], "UID")
+                if uid_values and uid_values[0] and isinstance(uid_values[0], bytes)
+                else "unknown"
+            )
+
+            if context:
+                logging.warning(
+                    "Failed to parse RRULE in time-range filter for %s (UID=%s): %s",
+                    context,
+                    uid,
+                    e,
+                )
+            else:
+                logging.warning(
+                    "Failed to parse RRULE in time-range filter (UID=%s): %s", uid, e
+                )
             return False
 
         # Get component handler for testing occurrences
@@ -816,17 +874,36 @@ class ComponentTimeRangeMatcher:
 
     def _get_occurrences_in_range(self, rrule, dtstart_parsed, query_start, query_end):
         """Generate RRULE occurrences within the specified time range."""
+        # Normalize query bounds to match the original DTSTART type/timezone
+        start_normalized = _normalize_dt_for_rrule(query_start, dtstart_parsed)
+        end_normalized = _normalize_dt_for_rrule(query_end, dtstart_parsed)
+
+        # For date-only events, extend the end bound to include the full day
         if isinstance(dtstart_parsed, date) and not isinstance(
             dtstart_parsed, datetime
         ):
-            # For date-only events, convert query bounds to datetime at midnight
-            query_start_dt = datetime.combine(query_start.date(), time())
-            query_end_dt = datetime.combine(query_end.date(), time()) + timedelta(
-                days=1
+            try:
+                end_normalized = end_normalized + timedelta(days=1)
+            except OverflowError:
+                # If we're already at the maximum date, keep end_normalized as-is
+                pass
+
+        # When the query end is unbounded (at MAX_EXPANSION_TIME), limit the number
+        # of instances to prevent resource exhaustion with infinite recurrences
+        # Compare dates to handle both aware and naive datetimes
+        query_end_date = (
+            query_end.date() if isinstance(query_end, datetime) else query_end
+        )
+        max_date = MAX_EXPANSION_TIME.date()
+        if query_end_date >= max_date:
+            # Use xafter with count limit for unbounded queries
+            occurrences = list(
+                rrule.xafter(start_normalized, count=MAX_RECURRENCE_INSTANCES, inc=True)
             )
-            return list(rrule.between(query_start_dt, query_end_dt, inc=True))
+            # Filter to only those before end_normalized
+            return [occ for occ in occurrences if occ <= end_normalized]
         else:
-            return list(rrule.between(query_start, query_end, inc=True))
+            return list(rrule.between(start_normalized, end_normalized, inc=True))
 
     def _test_occurrences(
         self,
@@ -894,6 +971,7 @@ class TextMatcher:
         text: str,
         collation: str | None = None,
         negate_condition: bool = False,
+        match_type: str = "contains",
     ) -> None:
         self.name = name
         self.type_fn = TYPES_FACTORY.for_property(name)
@@ -903,6 +981,7 @@ class TextMatcher:
             collation = "i;ascii-casemap"
         self.collation = _mod_collation.get_collation(collation)
         self.negate_condition = negate_condition
+        self.match_type = match_type
 
     def __repr__(self) -> str:
         return (
@@ -917,9 +996,9 @@ class TextMatcher:
 
     def match(self, prop: vText | vCategory | str):
         if isinstance(prop, vText):
-            matches = self.collation(str(prop), self.text, "contains")
+            matches = self.collation(str(prop), self.text, self.match_type)
         elif isinstance(prop, str):
-            matches = self.collation(prop, self.text, "contains")
+            matches = self.collation(prop, self.text, self.match_type)
         elif isinstance(prop, vCategory):
             matches = any([self.match(cat) for cat in prop.cats])
         else:
@@ -1018,17 +1097,7 @@ class ComponentFilter:
         # XML elements also match the targeted calendar component;
         for child in self.children:
             if isinstance(child, ComponentFilter):
-                # Special handling for VALARM components with time-range
-                if child.name == "VALARM" and child.time_range is not None:
-                    # Create enriched VALARM components with parent info
-                    subcomponents = []
-                    for c in comp.subcomponents:
-                        if c.name == "VALARM":
-                            # Create a copy with parent info attached
-                            enriched = _create_enriched_valarm(c, comp)
-                            subcomponents.append(enriched)
-                else:
-                    subcomponents = comp.subcomponents
+                subcomponents = self._get_subcomponents_for_matching(child, comp)
                 if not any(child.match(c, tzify) for c in subcomponents):
                     return False
             elif isinstance(child, PropertyFilter):
@@ -1039,12 +1108,40 @@ class ComponentFilter:
 
         return True
 
+    def _get_subcomponents_for_matching(
+        self, child_filter: "ComponentFilter", parent_comp: Component
+    ) -> list[Component]:
+        """Get subcomponents for matching, with special handling for VALARM.
+
+        When filtering VALARM components with time-range, we need to enrich them
+        by converting relative TRIGGER values to absolute times based on the parent
+        component's timing properties.
+        """
+        if child_filter.name == "VALARM" and child_filter.time_range is not None:
+            # Create enriched VALARM components with absolute trigger times
+            return [
+                _create_enriched_valarm(c, parent_comp)
+                for c in parent_comp.subcomponents
+                if c.name == "VALARM"
+            ]
+        else:
+            return parent_comp.subcomponents
+
     def _implicitly_defined(self):
         return any(
             not getattr(child, "is_not_defined", False) for child in self.children
         )
 
-    def match_indexes(self, indexes: IndexDict, tzify: TzifyFunction):
+    def match_indexes(
+        self, indexes: IndexDict, tzify: TzifyFunction, context: str | None = None
+    ):
+        """Match indexes against this component filter.
+
+        Args:
+            indexes: Index dictionary to match against
+            tzify: Timezone conversion function
+            context: Optional context string (e.g. filename) for error/warning reporting
+        """
         myindex = "C=" + self.name
         if self.is_not_defined:
             return not bool(indexes[myindex])
@@ -1052,12 +1149,12 @@ class ComponentFilter:
         subindexes = create_subindexes(indexes, myindex)
 
         if self.time_range is not None and not self.time_range.match_indexes(
-            subindexes, tzify
+            subindexes, tzify, context
         ):
             return False
 
         for child in self.children:
-            if not child.match_indexes(subindexes, tzify):
+            if not child.match_indexes(subindexes, tzify, context):
                 return False
 
         if not self._implicitly_defined():
@@ -1109,10 +1206,18 @@ class PropertyFilter:
         return self.time_range
 
     def filter_text_match(
-        self, text: str, collation: str | None = None, negate_condition: bool = False
+        self,
+        text: str,
+        collation: str | None = None,
+        negate_condition: bool = False,
+        match_type: str = "contains",
     ) -> TextMatcher:
         ret = TextMatcher(
-            self.name, text, collation=collation, negate_condition=negate_condition
+            self.name,
+            text,
+            collation=collation,
+            negate_condition=negate_condition,
+            match_type=match_type,
         )
         self.children.append(ret)
         return ret
@@ -1142,7 +1247,16 @@ class PropertyFilter:
 
         return True
 
-    def match_indexes(self, indexes: SubIndexDict, tzify: TzifyFunction) -> bool:
+    def match_indexes(
+        self, indexes: SubIndexDict, tzify: TzifyFunction, context: str | None = None
+    ) -> bool:
+        """Match indexes against this property filter.
+
+        Args:
+            indexes: Index dictionary to match against
+            tzify: Timezone conversion function
+            context: Optional context string (e.g. filename) for error/warning reporting
+        """
         myindex = "P=" + self.name
         if self.is_not_defined:
             return not bool(indexes[myindex])
@@ -1151,7 +1265,7 @@ class PropertyFilter:
             return bool(indexes[myindex])
 
         if self.time_range is not None and not self.time_range.match_indexes(
-            subindexes, tzify
+            subindexes, tzify, context
         ):
             return False
 
@@ -1185,10 +1299,18 @@ class ParameterFilter:
         self.children = children or []
 
     def filter_text_match(
-        self, text: str, collation: str | None = None, negate_condition: bool = False
+        self,
+        text: str,
+        collation: str | None = None,
+        negate_condition: bool = False,
+        match_type: str = "contains",
     ) -> TextMatcher:
         ret = TextMatcher(
-            self.name, text, collation=collation, negate_condition=negate_condition
+            self.name,
+            text,
+            collation=collation,
+            negate_condition=negate_condition,
+            match_type=match_type,
         )
         self.children.append(ret)
         return ret
@@ -1289,7 +1411,7 @@ class CalendarFilter(Filter):
     def check_from_indexes(self, name: str, indexes: IndexDict) -> bool:
         for child_filter in self.children:
             try:
-                if not child_filter.match_indexes(indexes, self.tzify):  # type: ignore
+                if not child_filter.match_indexes(indexes, self.tzify, name):  # type: ignore
                     return False
             except MissingProperty as e:
                 logging.warning(
@@ -1409,31 +1531,51 @@ class ICalendarFile(File):
         raise KeyError
 
     def _get_index(self, key: IndexKey) -> IndexValueIterator:
+        # Track parent component for VALARM enrichment
         # Use the original calendar without expansion for indexing
         # RRULE expansion will be handled at query time in match_indexes()
-        todo = [(self.calendar, key.split("/"))]
-        rest = []
+        todo: list[tuple[Component, list[str], Component | None]] = [
+            (self.calendar, key.split("/"), None)
+        ]
+        rest: list[tuple[Component, list[str], Component | None]] = []
         c: Component
         while todo:
-            (c, segments) = todo.pop(0)
+            (c, segments, parent) = todo.pop(0)
             if segments and segments[0].startswith("C="):
                 if c.name == segments[0][2:]:
                     if len(segments) > 1 and segments[1].startswith("C="):
-                        todo.extend((comp, segments[1:]) for comp in c.subcomponents)
+                        # Pass current component as parent for subcomponents
+                        todo.extend((comp, segments[1:], c) for comp in c.subcomponents)
                     else:
-                        rest.append((c, segments[1:]))
-        for c, segments in rest:
+                        rest.append((c, segments[1:], parent))
+
+        for c, segments, parent in rest:
             if not segments:
                 yield True
             elif segments[0].startswith("P="):
                 assert len(segments) == 1
+                prop_name = segments[0][2:]
                 try:
-                    p = c[segments[0][2:]]
+                    p = c[prop_name]
                 except KeyError:
                     pass
                 else:
                     if p is not None:
-                        yield p.to_ical()
+                        ical = p.to_ical()
+                        # Special handling for VALARM TRIGGER property
+                        if (
+                            c.name == "VALARM"
+                            and prop_name == "TRIGGER"
+                            and parent is not None
+                            and isinstance(p.dt, timedelta)
+                        ):
+                            # Create enriched VALARM to get absolute trigger time.
+                            # This ensures index values match what the filter will see,
+                            # preventing "index based filter not matching real file filter" errors.
+                            enriched = _create_enriched_valarm(c, parent)
+                            if "TRIGGER" in enriched:
+                                ical = enriched["TRIGGER"].to_ical()
+                        yield ical
             else:
                 raise AssertionError(f"segments: {segments!r}")
 
@@ -1452,20 +1594,180 @@ def as_tz_aware_ts(dt: datetime | date, default_timezone: str | timezone) -> dat
     return _dt
 
 
+@overload
+def _normalize_to_dtstart_type(
+    dt_value: date | datetime, dtstart: datetime
+) -> datetime: ...
+
+
+@overload
+def _normalize_to_dtstart_type(dt_value: date | datetime, dtstart: date) -> date: ...
+
+
+def _normalize_to_dtstart_type(
+    dt_value: date | datetime, dtstart: date | datetime
+) -> date | datetime:
+    """Normalize a date/datetime value to match the type of DTSTART.
+
+    This is necessary because dateutil.rrule cannot compare date and datetime objects.
+    When adding EXDATE/RDATE values to an rruleset, they must be the same type as DTSTART.
+
+    Args:
+        dt_value: The date or datetime value to normalize
+        dtstart: The DTSTART value to match the type of
+
+    Returns:
+        The normalized value matching DTSTART's type
+    """
+    # If both are the same type, check timezone compatibility
+    if type(dt_value) is type(dtstart):
+        # Both are datetimes - ensure timezone awareness matches
+        if isinstance(dt_value, datetime) and isinstance(dtstart, datetime):
+            if dt_value.tzinfo is not None and dtstart.tzinfo is None:
+                # Converting aware to naive - extract date and use dtstart's time
+                return datetime.combine(dt_value.date(), dtstart.time())
+            elif dt_value.tzinfo is None and dtstart.tzinfo is not None:
+                # Make dt_value aware to match dtstart
+                return dt_value.replace(tzinfo=dtstart.tzinfo)
+        return dt_value
+
+    # If DTSTART is a date (not datetime), convert datetime to date
+    if isinstance(dtstart, date) and not isinstance(dtstart, datetime):
+        if isinstance(dt_value, datetime):
+            return dt_value.date()
+        return dt_value
+
+    # If DTSTART is a datetime, convert date to datetime
+    if isinstance(dtstart, datetime):
+        if isinstance(dt_value, date) and not isinstance(dt_value, datetime):
+            # Convert date to datetime, matching DTSTART's time and timezone
+            # This ensures that EXDATE;VALUE=DATE:20240102 matches the occurrence
+            # at DTSTART's time (e.g., 10:00:00), not just midnight
+            if dtstart.tzinfo is not None:
+                # Use the same time and timezone as DTSTART
+                return datetime.combine(dt_value, dtstart.time()).replace(
+                    tzinfo=dtstart.tzinfo
+                )
+            else:
+                # Naive datetime - use DTSTART's time
+                return datetime.combine(dt_value, dtstart.time())
+        # dt_value is already a datetime, but check timezone awareness
+        if isinstance(dt_value, datetime):
+            if dt_value.tzinfo is not None and dtstart.tzinfo is None:
+                # Converting aware datetime to naive datetime
+                # First extract the date to handle timezone properly, then
+                # combine with dtstart's time to create a naive datetime
+                return datetime.combine(dt_value.date(), dtstart.time())
+            elif dt_value.tzinfo is None and dtstart.tzinfo is not None:
+                # Make dt_value aware to match dtstart
+                return dt_value.replace(tzinfo=dtstart.tzinfo)
+        return dt_value
+
+    return dt_value
+
+
+def _normalize_rrule_until(rrule_str: str, dtstart: date | datetime) -> str:
+    """Normalize RRULE UNTIL to be compatible with dateutil when DTSTART is timezone-aware.
+
+    RFC 5545 requires UNTIL to be in UTC when DTSTART is timezone-aware.
+    Some clients (like iPhone) send UNTIL without UTC suffix.
+    This function parses the RRULE and ensures UNTIL is properly formatted.
+
+    Args:
+        rrule_str: The RRULE string
+        dtstart: The DTSTART value (used to determine if normalization is needed)
+
+    Returns:
+        The RRULE string, possibly modified to have UNTIL in UTC format
+    """
+    # Only normalize if dtstart is a timezone-aware datetime
+    if not isinstance(dtstart, datetime) or dtstart.tzinfo is None:
+        return rrule_str
+
+    parsed = vRecur.from_ical(rrule_str)
+    until_list = parsed.get("UNTIL", [])
+
+    if not until_list:
+        return rrule_str
+
+    until = until_list[0]
+
+    # If UNTIL is a date (not datetime), convert to end of day in UTC
+    if isinstance(until, date) and not isinstance(until, datetime):
+        until_dt = datetime.combine(until, time(23, 59, 59), tzinfo=timezone.utc)
+        parsed["UNTIL"] = [until_dt]
+        return parsed.to_ical().decode("utf-8")
+
+    # If UNTIL is a naive datetime, make it UTC (issue #80, #571)
+    if isinstance(until, datetime) and until.tzinfo is None:
+        parsed["UNTIL"] = [until.replace(tzinfo=timezone.utc)]
+        return parsed.to_ical().decode("utf-8")
+
+    return rrule_str
+
+
 def rruleset_from_comp(comp: Component) -> dateutil.rrule.rruleset:
     dtstart = comp["DTSTART"].dt
     rrulestr = comp["RRULE"].to_ical().decode("utf-8")
+    # Normalize UNTIL to be UTC if dtstart is timezone-aware
+    rrulestr = _normalize_rrule_until(rrulestr, dtstart)
     rrule = dateutil.rrule.rrulestr(rrulestr, dtstart=dtstart)
     rs = dateutil.rrule.rruleset()
     rs.rrule(rrule)  # type: ignore
+
+    # dateutil.rrule internally converts date objects to datetime objects.
+    # To determine what type the rrule will generate, we check the first occurrence.
+    # This is necessary because EXDATE/RDATE values must match the generated type.
+    first_occurrence = next(iter(rrule), None)
+    effective_dtstart: datetime
+    if first_occurrence is not None:
+        effective_dtstart = first_occurrence
+    elif isinstance(dtstart, date) and not isinstance(dtstart, datetime):
+        # dateutil.rrule converts date to datetime at midnight
+        effective_dtstart = datetime.combine(dtstart, datetime.min.time())
+    else:
+        # dtstart must be a datetime at this point
+        assert isinstance(dtstart, datetime)
+        effective_dtstart = dtstart
+
     if "EXDATE" in comp:
-        for exdate in comp["EXDATE"]:
-            rs.exdate(exdate)
+        exdate_prop = comp["EXDATE"]
+        # EXDATE can be either:
+        # 1. A single vDDDLists (one EXDATE property with one or more dates)
+        # 2. A list of vDDDLists (multiple EXDATE properties)
+        # Extract the actual datetime/date values from the .dts list(s)
+        # and normalize them to match what the rrule will generate
+        if isinstance(exdate_prop, list):
+            for exdate_list in exdate_prop:
+                for exdate in exdate_list.dts:
+                    normalized = _normalize_to_dtstart_type(
+                        exdate.dt, effective_dtstart
+                    )
+                    rs.exdate(normalized)
+        else:
+            for exdate in exdate_prop.dts:
+                normalized = _normalize_to_dtstart_type(exdate.dt, effective_dtstart)
+                rs.exdate(normalized)
     if "RDATE" in comp:
-        for rdate in comp["RDATE"]:
-            rs.rdate(rdate)
+        rdate_prop = comp["RDATE"]
+        # RDATE can be either:
+        # 1. A single vDDDLists (one RDATE property with one or more dates)
+        # 2. A list of vDDDLists (multiple RDATE properties)
+        # Extract the actual datetime/date values from the .dts list(s)
+        # and normalize them to match what the rrule will generate
+        if isinstance(rdate_prop, list):
+            for rdate_list in rdate_prop:
+                for rdate in rdate_list.dts:
+                    normalized = _normalize_to_dtstart_type(rdate.dt, effective_dtstart)
+                    rs.rdate(normalized)
+        else:
+            for rdate in rdate_prop.dts:
+                normalized = _normalize_to_dtstart_type(rdate.dt, effective_dtstart)
+                rs.rdate(normalized)
     if "EXRULE" in comp:
         exrulestr = comp["EXRULE"].to_ical().decode("utf-8")
+        # Normalize UNTIL to be UTC if dtstart is timezone-aware
+        exrulestr = _normalize_rrule_until(exrulestr, dtstart)
         exrule = dateutil.rrule.rrulestr(exrulestr, dtstart=dtstart)
         assert isinstance(exrule, dateutil.rrule.rrule)
         rs.exrule(exrule)
@@ -1591,7 +1893,20 @@ def _expand_rrule_component(
         start_normalized = _normalize_dt_for_rrule(adjusted_start, original_dtstart.dt)
         end_normalized = _normalize_dt_for_rrule(end, original_dtstart.dt)
 
-        occurrences = rs.between(start_normalized, end_normalized, inc=True)
+        # When the query end is unbounded (at MAX_EXPANSION_TIME), limit the number
+        # of instances to prevent resource exhaustion with infinite recurrences
+        # Compare dates to handle both aware and naive datetimes
+        end_date = end.date() if isinstance(end, datetime) else end
+        max_date = MAX_EXPANSION_TIME.date()
+        if end_date >= max_date:
+            # Use xafter with count limit for unbounded queries
+            all_occurrences = list(
+                rs.xafter(start_normalized, count=MAX_RECURRENCE_INSTANCES, inc=True)
+            )
+            # Filter to only those before end_normalized
+            occurrences = [occ for occ in all_occurrences if occ <= end_normalized]
+        else:
+            occurrences = rs.between(start_normalized, end_normalized, inc=True)
     else:
         # For unbounded queries, we still need to return a list/iterator
         occurrences = rs  # type: ignore
@@ -1823,6 +2138,92 @@ def limit_calendar_recurrence_set(
     return outcal
 
 
+def limit_calendar_freebusy_set(
+    incal: Calendar, start: datetime, end: datetime
+) -> Calendar:
+    """Limit FREEBUSY properties to a specified time range.
+
+    Filters FREEBUSY properties in VFREEBUSY components to only include
+    those that overlap with the specified time range. This implements
+    RFC 4791 section 9.6.7.
+
+    Args:
+        incal: Input calendar
+        start: Start of time range (UTC)
+        end: End of time range (UTC)
+
+    Returns:
+        Calendar with VFREEBUSY components containing only overlapping FREEBUSY periods
+    """
+    outcal = Calendar()
+    if incal.name != "VCALENDAR":
+        raise AssertionError(f"called on file with root component {incal.name}")
+
+    # Copy calendar properties
+    for field in incal:
+        outcal[field] = incal[field]
+
+    # Normalize start/end to naive UTC for comparison
+    start_utc = asutc(start) if isinstance(start, datetime) else start
+    end_utc = asutc(end) if isinstance(end, datetime) else end
+
+    # First, add all VTIMEZONE components to preserve timezone definitions
+    for insub in incal.subcomponents:
+        if insub.name == "VTIMEZONE":
+            outcal.add_component(insub)
+
+    # Process other components
+    for insub in incal.subcomponents:
+        if insub.name == "VTIMEZONE":
+            continue
+
+        if insub.name == "VFREEBUSY":
+            # Create a new VFREEBUSY component with filtered FREEBUSY properties
+            from icalendar.cal import FreeBusy
+
+            newsub = FreeBusy()
+
+            # Copy all non-FREEBUSY properties
+            for field in insub:
+                if field != "FREEBUSY":
+                    newsub[field] = insub[field]
+
+            # Filter and copy only overlapping FREEBUSY properties
+            freebusy_props = insub.get("FREEBUSY", [])
+            if not isinstance(freebusy_props, list):
+                freebusy_props = [freebusy_props]
+
+            for fb_period in freebusy_props:
+                # FREEBUSY properties are vPeriod objects with start and end times
+                # Check if the period overlaps with the requested range
+                # Overlap occurs if: period_start < range_end AND period_end > range_start
+                period_start_utc = (
+                    asutc(fb_period.start)
+                    if isinstance(fb_period.start, datetime)
+                    else fb_period.start
+                )
+                period_end_utc = (
+                    asutc(fb_period.end)
+                    if isinstance(fb_period.end, datetime)
+                    else fb_period.end
+                )
+
+                if period_start_utc < end_utc and period_end_utc > start_utc:
+                    # This FREEBUSY period overlaps with the requested range
+                    newsub.add("FREEBUSY", fb_period)
+
+            # Copy subcomponents (though VFREEBUSY typically doesn't have any)
+            for subcomp in insub.subcomponents:
+                newsub.add_component(subcomp)
+
+            outcal.add_component(newsub)
+        else:
+            # For non-VFREEBUSY components, just copy them as-is
+            outcal.add_component(insub)
+
+    return outcal
+
+
 def asutc(dt):
     if isinstance(dt, date) and not isinstance(dt, datetime):
         # Return date as-is - dates are timezone-agnostic
