@@ -21,19 +21,20 @@
 
 import configparser
 import errno
+import functools
 import logging
 import os
 import shutil
 import stat
 import uuid
 from io import BytesIO, StringIO
-from typing import Optional, Iterable
+from typing import cast
+from collections.abc import Iterable
 
 import dulwich.repo
-from dulwich.file import FileLocked, GitFile
-from dulwich.index import Index, index_entry_from_stat, write_index_dict
-from dulwich.objects import Blob, Tree
-from dulwich.pack import SHA1Writer
+from dulwich.file import FileLocked
+from dulwich.index import IndexEntry, index_entry_from_stat, locked_index
+from dulwich.objects import Blob, Commit, Tree
 
 from . import (
     DEFAULT_MIME_TYPE,
@@ -51,8 +52,8 @@ from . import (
     open_by_content_type,
     open_by_extension,
 )
-from .config import FILENAME as CONFIG_FILENAME
-from .config import CollectionMetadata, FileBasedCollectionMetadata
+from .config import CONFIG_FILENAME
+from .config import CollectionMetadata, FileBasedCollectionMetadata, is_metadata_file
 from .index import MemoryIndex
 
 DEFAULT_ENCODING = "utf-8"
@@ -82,8 +83,7 @@ class RepoCollectionMetadata(CollectionMetadata):
         if url is not None:
             config.set(b"xandikos", b"source", url.encode(DEFAULT_ENCODING))
         else:
-            # TODO(jelmer): Add and use config.remove()
-            config.set(b"xandikos", b"source", b"")
+            config.remove(b"xandikos", b"source")
         self._write_config(config)
 
     def get_color(self):
@@ -98,8 +98,7 @@ class RepoCollectionMetadata(CollectionMetadata):
         if color is not None:
             config.set(b"xandikos", b"color", color.encode(DEFAULT_ENCODING))
         else:
-            # TODO(jelmer): Add and use config.remove()
-            config.set(b"xandikos", b"color", b"")
+            config.remove(b"xandikos", b"color")
         self._write_config(config)
 
     def _write_config(self, config):
@@ -123,7 +122,7 @@ class RepoCollectionMetadata(CollectionMetadata):
                 displayname.encode(DEFAULT_ENCODING),
             )
         else:
-            config.set(b"xandikos", b"displayname", b"")
+            config.remove(b"xandikos", b"displayname")
         self._write_config(config)
 
     def get_description(self):
@@ -150,8 +149,7 @@ class RepoCollectionMetadata(CollectionMetadata):
         if comment is not None:
             config.set(b"xandikos", b"comment", comment.encode(DEFAULT_ENCODING))
         else:
-            # TODO(jelmer): Add and use config.remove()
-            config.set(b"xandikos", b"comment", b"")
+            config.remove(b"xandikos", b"comment")
         self._write_config(config)
 
     def set_type(self, store_type):
@@ -181,27 +179,37 @@ class RepoCollectionMetadata(CollectionMetadata):
         config.set(b"xandikos", b"calendar-order", order.encode("utf-8"))
         self._write_config(config)
 
+    def get_refreshrate(self):
+        config = self._repo.get_config()
+        refreshrate = config.get(b"xandikos", b"refreshrate")
+        if refreshrate == b"":
+            raise KeyError
+        return refreshrate.decode(DEFAULT_ENCODING)
 
-class locked_index:
-    def __init__(self, path) -> None:
-        self._path = path
-
-    def __enter__(self):
-        self._file = GitFile(self._path, "wb")
-        self._index = Index(self._path)
-        return self._index
-
-    def __exit__(self, exc_type, exc_value, traceback):
-        if exc_type is not None:
-            self._file.abort()
-            return
-        try:
-            f = SHA1Writer(self._file)
-            write_index_dict(f, self._index._byname)
-        except BaseException:
-            self._file.abort()
+    def set_refreshrate(self, refreshrate):
+        config = self._repo.get_config()
+        if refreshrate is not None:
+            config.set(
+                b"xandikos", b"refreshrate", refreshrate.encode(DEFAULT_ENCODING)
+            )
         else:
-            f.close()
+            config.remove(b"xandikos", b"refreshrate")
+        self._write_config(config)
+
+    def get_timezone(self):
+        config = self._repo.get_config()
+        timezone = config.get(b"xandikos", b"timezone")
+        if timezone == b"":
+            raise KeyError
+        return timezone.decode(DEFAULT_ENCODING)
+
+    def set_timezone(self, timezone):
+        config = self._repo.get_config()
+        if timezone is not None:
+            config.set(b"xandikos", b"timezone", timezone.encode(DEFAULT_ENCODING))
+        else:
+            config.remove(b"xandikos", b"timezone")
+        self._write_config(config)
 
 
 class GitStore(Store):
@@ -226,6 +234,37 @@ class GitStore(Store):
         # Set of blob ids that have already been scanned
         self._fname_to_uid: dict[str, tuple[str, str]] = {}
 
+        # Cache parsed files by blob SHA - avoids reparsing identical content
+        self._parsed_file_cache = functools.lru_cache(maxsize=100)(
+            self._parse_file_by_sha
+        )
+
+    def _parse_file_by_sha(self, sha: str, content_type: str | None, name: str):
+        """Parse a file by its SHA, used for caching."""
+        blob = self.repo.object_store[sha.encode("ascii")]
+        if content_type is None:
+            return open_by_extension(
+                blob.chunked,
+                name,
+                extra_file_handlers=self.extra_file_handlers,
+            )
+        else:
+            return open_by_content_type(
+                blob.chunked,
+                content_type,
+                extra_file_handlers=self.extra_file_handlers,
+            )
+
+    def get_file(
+        self, name: str, content_type: str | None = None, etag: str | None = None
+    ):
+        """Get file with caching based on blob SHA."""
+        if etag is None:
+            etag = self._get_etag(name)
+
+        # Use cached parsing based on blob SHA
+        return self._parsed_file_cache(etag, content_type, name)
+
     def _get_etag(self, name: str) -> str:
         raise NotImplementedError(self._get_etag)
 
@@ -234,7 +273,7 @@ class GitStore(Store):
         name: str,
         data: Iterable[bytes],
         message: str,
-        author: Optional[str] = None,
+        author: str | None = None,
     ):
         raise NotImplementedError(self._import_one)
 
@@ -292,9 +331,10 @@ class GitStore(Store):
         name: str,
         content_type: str,
         data: Iterable[bytes],
-        message: Optional[str] = None,
-        author: Optional[str] = None,
-        replace_etag: Optional[str] = None,
+        message: str | None = None,
+        author: str | None = None,
+        replace_etag: str | None = None,
+        requester: str | None = None,
     ) -> tuple[str, str]:
         """Import a single object.
 
@@ -305,6 +345,7 @@ class GitStore(Store):
           message: Commit message
           author: Optional author
           replace_etag: optional etag of object to replace
+          requester: Optional User-Agent or client information
         Raises:
           InvalidETag: when the name already exists but with different etag
           DuplicateUidError: when the uid already exists
@@ -331,6 +372,11 @@ class GitStore(Store):
             except KeyError:
                 old_fi = None
             message = "\n".join(fi.describe_delta(name, old_fi))
+
+        # Add requester information to commit message in RFC822 style
+        if requester is not None:
+            message = message + "\n\nRequester: " + requester
+
         etag = self._import_one(name, fi.normalized(), message, author=author)
         return (name, etag.decode("ascii"))
 
@@ -362,8 +408,8 @@ class GitStore(Store):
             except KeyError:
                 logger.warning("No UID found in file %s", name)
                 uid = None
-            except InvalidFileContents:
-                logging.warning("Unable to parse file %s", name)
+            except InvalidFileContents as e:
+                logging.warning("Unable to parse file %s: %s", name, e)
                 uid = None
             except NotImplementedError:
                 # This file type doesn't support UIDs
@@ -593,7 +639,7 @@ class BareGitStore(GitStore):
                 raise InvalidCTag(ctag) from exc
         for name, mode, sha in tree.iteritems():
             name = name.decode(DEFAULT_ENCODING)
-            if name == CONFIG_FILENAME:
+            if is_metadata_file(name):
                 continue
             yield (name, mode, sha)
 
@@ -607,16 +653,54 @@ class BareGitStore(GitStore):
         return cls(repo)
 
     def _commit_tree(self, tree_id, message, author=None):
-        return self.repo.do_commit(
-            message=message, tree=tree_id, ref=self.ref, author=author
-        )
+        """Create a commit for the given tree.
+
+        Args:
+            tree_id: Tree object ID
+            message: Commit message (bytes)
+            author: Optional author (bytes)
+
+        Returns:
+            Commit SHA
+        """
+        import time
+        from dulwich.porcelain import get_user_identity
+
+        c = Commit()
+        c.tree = tree_id
+
+        if author is None:
+            author = get_user_identity(self.repo.get_config_stack())
+
+        c.author = author
+        c.committer = author
+        c.author_time = int(time.time())
+        c.commit_time = c.author_time
+        c.author_timezone = 0
+        c.commit_timezone = 0
+        c.encoding = b"UTF-8"
+        c.message = message
+
+        # Get parent commits
+        try:
+            c.parents = [self.repo.refs[self.ref]]
+        except KeyError:
+            c.parents = []
+
+        # Add commit to object store
+        self.repo.object_store.add_object(c)
+
+        # Update ref
+        self.repo.refs[self.ref] = c.id
+
+        return c.id
 
     def _import_one(
         self,
         name: str,
         data: Iterable[bytes],
         message: str,
-        author: Optional[str] = None,
+        author: str | None = None,
     ) -> bytes:
         """Import a single object.
 
@@ -648,7 +732,7 @@ class BareGitStore(GitStore):
           etag: Optional mandatory etag of object to remove
         Raises:
           NoSuchItem: when the item doesn't exist
-          InvalidETag: If the specified ETag doesn't match the curren
+          InvalidETag: If the specified ETag doesn't match the current
         """
         tree = self._get_current_tree()
         name_enc = name.encode(DEFAULT_ENCODING)
@@ -711,14 +795,16 @@ class TreeGitStore(GitStore):
 
     def _commit_tree(self, index, message, author=None):
         tree = index.commit(self.repo.object_store)
-        return self.repo.do_commit(message=message, author=author, tree=tree)
+        return self.repo.get_worktree().commit(
+            message=message, author=author, tree=tree
+        )
 
     def _import_one(
         self,
         name: str,
         data: Iterable[bytes],
         message: str,
-        author: Optional[str] = None,
+        author: str | None = None,
     ) -> bytes:
         """Import a single object.
 
@@ -732,12 +818,22 @@ class TreeGitStore(GitStore):
         try:
             with locked_index(self.repo.index_path()) as index:
                 p = os.path.join(self.repo.path, name)
+                # Create directory if it doesn't exist
+                os.makedirs(os.path.dirname(p), exist_ok=True)
                 with open(p, "wb") as f:
                     f.writelines(data)
                 st = os.lstat(p)
-                blob = Blob.from_string(b"".join(data))
+                # Read the file back to create the blob since data iterator is consumed
+                with open(p, "rb") as f:
+                    blob = Blob.from_string(f.read())
                 encoded_name = name.encode(DEFAULT_ENCODING)
-                if encoded_name not in index or blob.id != index[encoded_name].sha:
+                try:
+                    entry = index[encoded_name]
+                except KeyError:
+                    entry = None
+                if entry is None or (
+                    isinstance(entry, IndexEntry) and blob.id != entry.sha
+                ):
                     self.repo.object_store.add_object(blob)
                     index[encoded_name] = index_entry_from_stat(st, blob.id)
                     self._commit_tree(
@@ -751,7 +847,13 @@ class TreeGitStore(GitStore):
                 raise OutOfSpaceError() from exc
             raise
 
-    def delete_one(self, name, message=None, author=None, etag=None):
+    def delete_one(
+        self,
+        name: str,
+        message: str | None = None,
+        author: str | None = None,
+        etag: str | None = None,
+    ) -> None:
         """Delete an item.
 
         Args:
@@ -761,12 +863,13 @@ class TreeGitStore(GitStore):
           etag: Optional mandatory etag of object to remove
         Raise:
           NoSuchItem: when the item doesn't exist
-          InvalidETag: If the specified ETag doesn't match the curren
+          InvalidETag: If the specified ETag doesn't match the current
         """
         p = os.path.join(self.repo.path, name)
         try:
             with open(p, "rb") as f:
-                current_blob = Blob.from_string(f.read())
+                # cast to Blob, because on dulwich < 0.24.0 Blob.from_string returns ShaFile
+                current_blob = cast(Blob, Blob.from_string(f.read()))
         except FileNotFoundError as exc:
             raise NoSuchItem(name) from exc
         except IsADirectoryError as exc:
@@ -789,7 +892,7 @@ class TreeGitStore(GitStore):
         except FileLocked:
             raise LockedError(name)
 
-    def get_ctag(self):
+    def get_ctag(self) -> str:
         """Return the ctag for this store."""
         index = self.repo.open_index()
         return index.commit(self.repo.object_store).decode("ascii")
@@ -806,14 +909,14 @@ class TreeGitStore(GitStore):
                 raise InvalidCTag(ctag) from exc
             for name, mode, sha in tree.iteritems():
                 name = name.decode(DEFAULT_ENCODING)
-                if name == CONFIG_FILENAME:
+                if is_metadata_file(name):
                     continue
                 yield (name, mode, sha)
         else:
             index = self.repo.open_index()
             for name, sha, mode in index.iterobjects():
                 name = name.decode(DEFAULT_ENCODING)
-                if name == CONFIG_FILENAME:
+                if is_metadata_file(name):
                     continue
                 yield (name, mode, sha)
 
@@ -824,7 +927,7 @@ class TreeGitStore(GitStore):
         """
         ret = []
         for name in os.listdir(self.path):
-            if name == dulwich.repo.CONTROLDIR:
+            if name == dulwich.repo.CONTROLDIR or is_metadata_file(name):
                 continue
             p = os.path.join(self.path, name)
             if os.path.isdir(p):
